@@ -9,202 +9,71 @@
 #endif
 
 #include "pycore_call.h"          // _PyObject_CallNoArgs()
+#include "pycore_dict.h"          // _PyDict_SizeOf()
 #include <ffi.h>
 #ifdef MS_WIN32
 #  include <malloc.h>
 #endif
 #include "ctypes.h"
 
-/******************************************************************/
-/*
-  StdDict - a dictionary subclass, containing additional C accessible fields
-
-  XXX blabla more
-*/
-
-/* Seems we need this, otherwise we get problems when calling
- * PyDict_SetItem() (ma_lookup is NULL)
+/* This file relates to StgInfo -- type-specific information for ctypes.
+ * See ctypes.h for details.
  */
-static int
-PyCStgDict_init(StgDictObject *self, PyObject *args, PyObject *kwds)
-{
-    if (PyDict_Type.tp_init((PyObject *)self, args, kwds) < 0)
-        return -1;
-    self->format = NULL;
-    self->ndim = 0;
-    self->shape = NULL;
-    return 0;
-}
-
-static int
-PyCStgDict_clear(StgDictObject *self)
-{
-    Py_CLEAR(self->proto);
-    Py_CLEAR(self->argtypes);
-    Py_CLEAR(self->converters);
-    Py_CLEAR(self->restype);
-    Py_CLEAR(self->checker);
-    return 0;
-}
-
-static void
-PyCStgDict_dealloc(StgDictObject *self)
-{
-    PyCStgDict_clear(self);
-    PyMem_Free(self->format);
-    PyMem_Free(self->shape);
-    PyMem_Free(self->ffi_type_pointer.elements);
-    PyDict_Type.tp_dealloc((PyObject *)self);
-}
-
-static PyObject *
-PyCStgDict_sizeof(StgDictObject *self, void *unused)
-{
-    Py_ssize_t res;
-
-    res = _PyDict_SizeOf((PyDictObject *)self);
-    res += sizeof(StgDictObject) - sizeof(PyDictObject);
-    if (self->format)
-        res += strlen(self->format) + 1;
-    res += self->ndim * sizeof(Py_ssize_t);
-    if (self->ffi_type_pointer.elements)
-        res += (self->length + 1) * sizeof(ffi_type *);
-    return PyLong_FromSsize_t(res);
-}
 
 int
-PyCStgDict_clone(StgDictObject *dst, StgDictObject *src)
+PyCStgInfo_clone(StgInfo *dst_info, StgInfo *src_info)
 {
-    char *d, *s;
     Py_ssize_t size;
 
-    PyCStgDict_clear(dst);
-    PyMem_Free(dst->ffi_type_pointer.elements);
-    PyMem_Free(dst->format);
-    dst->format = NULL;
-    PyMem_Free(dst->shape);
-    dst->shape = NULL;
-    dst->ffi_type_pointer.elements = NULL;
+    ctype_clear_stginfo(dst_info);
+    PyMem_Free(dst_info->ffi_type_pointer.elements);
+    PyMem_Free(dst_info->format);
+    dst_info->format = NULL;
+    PyMem_Free(dst_info->shape);
+    dst_info->shape = NULL;
+    dst_info->ffi_type_pointer.elements = NULL;
 
-    d = (char *)dst;
-    s = (char *)src;
-    memcpy(d + sizeof(PyDictObject),
-           s + sizeof(PyDictObject),
-           sizeof(StgDictObject) - sizeof(PyDictObject));
+    memcpy(dst_info, src_info, sizeof(StgInfo));
 
-    Py_XINCREF(dst->proto);
-    Py_XINCREF(dst->argtypes);
-    Py_XINCREF(dst->converters);
-    Py_XINCREF(dst->restype);
-    Py_XINCREF(dst->checker);
+    Py_XINCREF(dst_info->proto);
+    Py_XINCREF(dst_info->argtypes);
+    Py_XINCREF(dst_info->converters);
+    Py_XINCREF(dst_info->restype);
+    Py_XINCREF(dst_info->checker);
+    Py_XINCREF(dst_info->module);
 
-    if (src->format) {
-        dst->format = PyMem_Malloc(strlen(src->format) + 1);
-        if (dst->format == NULL) {
+    if (src_info->format) {
+        dst_info->format = PyMem_Malloc(strlen(src_info->format) + 1);
+        if (dst_info->format == NULL) {
             PyErr_NoMemory();
             return -1;
         }
-        strcpy(dst->format, src->format);
+        strcpy(dst_info->format, src_info->format);
     }
-    if (src->shape) {
-        dst->shape = PyMem_Malloc(sizeof(Py_ssize_t) * src->ndim);
-        if (dst->shape == NULL) {
+    if (src_info->shape) {
+        dst_info->shape = PyMem_Malloc(sizeof(Py_ssize_t) * src_info->ndim);
+        if (dst_info->shape == NULL) {
             PyErr_NoMemory();
             return -1;
         }
-        memcpy(dst->shape, src->shape,
-               sizeof(Py_ssize_t) * src->ndim);
+        memcpy(dst_info->shape, src_info->shape,
+               sizeof(Py_ssize_t) * src_info->ndim);
     }
 
-    if (src->ffi_type_pointer.elements == NULL)
+    if (src_info->ffi_type_pointer.elements == NULL)
         return 0;
-    size = sizeof(ffi_type *) * (src->length + 1);
-    dst->ffi_type_pointer.elements = PyMem_Malloc(size);
-    if (dst->ffi_type_pointer.elements == NULL) {
+    size = sizeof(ffi_type *) * (src_info->length + 1);
+    dst_info->ffi_type_pointer.elements = PyMem_Malloc(size);
+    if (dst_info->ffi_type_pointer.elements == NULL) {
         PyErr_NoMemory();
         return -1;
     }
-    memcpy(dst->ffi_type_pointer.elements,
-           src->ffi_type_pointer.elements,
+    memcpy(dst_info->ffi_type_pointer.elements,
+           src_info->ffi_type_pointer.elements,
            size);
     return 0;
 }
 
-static struct PyMethodDef PyCStgDict_methods[] = {
-    {"__sizeof__", (PyCFunction)PyCStgDict_sizeof, METH_NOARGS},
-    {NULL, NULL}                /* sentinel */
-};
-
-PyTypeObject PyCStgDict_Type = {
-    PyVarObject_HEAD_INIT(NULL, 0)
-    "StgDict",
-    sizeof(StgDictObject),
-    0,
-    (destructor)PyCStgDict_dealloc,             /* tp_dealloc */
-    0,                                          /* tp_vectorcall_offset */
-    0,                                          /* tp_getattr */
-    0,                                          /* tp_setattr */
-    0,                                          /* tp_as_async */
-    0,                                          /* tp_repr */
-    0,                                          /* tp_as_number */
-    0,                                          /* tp_as_sequence */
-    0,                                          /* tp_as_mapping */
-    0,                                          /* tp_hash */
-    0,                                          /* tp_call */
-    0,                                          /* tp_str */
-    0,                                          /* tp_getattro */
-    0,                                          /* tp_setattro */
-    0,                                          /* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /* tp_flags */
-    0,                                          /* tp_doc */
-    0,                                          /* tp_traverse */
-    0,                                          /* tp_clear */
-    0,                                          /* tp_richcompare */
-    0,                                          /* tp_weaklistoffset */
-    0,                                          /* tp_iter */
-    0,                                          /* tp_iternext */
-    PyCStgDict_methods,                         /* tp_methods */
-    0,                                          /* tp_members */
-    0,                                          /* tp_getset */
-    0,                                          /* tp_base */
-    0,                                          /* tp_dict */
-    0,                                          /* tp_descr_get */
-    0,                                          /* tp_descr_set */
-    0,                                          /* tp_dictoffset */
-    (initproc)PyCStgDict_init,                          /* tp_init */
-    0,                                          /* tp_alloc */
-    0,                                          /* tp_new */
-    0,                                          /* tp_free */
-};
-
-/* May return NULL, but does not set an exception! */
-StgDictObject *
-PyType_stgdict(PyObject *obj)
-{
-    PyTypeObject *type;
-
-    if (!PyType_Check(obj))
-        return NULL;
-    type = (PyTypeObject *)obj;
-    if (!type->tp_dict || !PyCStgDict_CheckExact(type->tp_dict))
-        return NULL;
-    return (StgDictObject *)type->tp_dict;
-}
-
-/* May return NULL, but does not set an exception! */
-/*
-  This function should be as fast as possible, so we don't call PyType_stgdict
-  above but inline the code, and avoid the PyType_Check().
-*/
-StgDictObject *
-PyObject_stgdict(PyObject *self)
-{
-    PyTypeObject *type = Py_TYPE(self);
-    if (!type->tp_dict || !PyCStgDict_CheckExact(type->tp_dict))
-        return NULL;
-    return (StgDictObject *)type->tp_dict;
-}
-
 /* descr is the descriptor for a field marked as anonymous.  Get all the
  _fields_ descriptors from descr->proto, create new descriptors with offset
  and index adjusted, and stuff them into type.
@@ -225,7 +94,7 @@ MakeFields(PyObject *type, CFieldObject *descr,
     if (fieldlist == NULL)
         return -1;
 
-    ctypes_state *st = GLOBAL_STATE();
+    ctypes_state *st = get_module_state_by_class(Py_TYPE(descr));
     PyTypeObject *cfield_tp = st->PyCField_Type;
     for (i = 0; i < PySequence_Fast_GET_SIZE(fieldlist); ++i) {
         PyObject *pair = PySequence_Fast_GET_ITEM(fieldlist, i); /* borrowed */
@@ -295,7 +164,7 @@ MakeAnonFields(PyObject *type)
     PyObject *anon_names;
     Py_ssize_t i;
 
-    if (_PyObject_LookupAttr(type, &_Py_ID(_anonymous_), &anon) < 0) {
+    if (PyObject_GetOptionalAttr(type, &_Py_ID(_anonymous_), &anon) < 0) {
         return -1;
     }
     if (anon == NULL) {
@@ -306,7 +175,7 @@ MakeAnonFields(PyObject *type)
     if (anon_names == NULL)
         return -1;
 
-    ctypes_state *st = GLOBAL_STATE();
+    ctypes_state *st = get_module_state_by_def(Py_TYPE(type));
     PyTypeObject *cfield_tp = st->PyCField_Type;
     for (i = 0; i < PySequence_Fast_GET_SIZE(anon_names); ++i) {
         PyObject *fname = PySequence_Fast_GET_ITEM(anon_names, i); /* borrowed */
@@ -366,18 +235,18 @@ _ctypes_alloc_format_padding(const char *prefix, Py_ssize_t padding)
 
 /*
   Retrieve the (optional) _pack_ attribute from a type, the _fields_ attribute,
-  and create an StgDictObject.  Used for Structure and Union subclasses.
+  and initialize StgInfo.  Used for Structure and Union subclasses.
 */
 int
-PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct)
+PyCStructUnionType_update_stginfo(PyObject *type, PyObject *fields, int isStruct)
 {
-    StgDictObject *stgdict, *basedict;
     Py_ssize_t len, offset, size, align, i;
     Py_ssize_t union_size, total_align, aligned_size;
     Py_ssize_t field_size = 0;
     int bitofs;
     PyObject *tmp;
     int pack;
+    int forced_alignment = 1;
     Py_ssize_t ffi_ofs;
     int big_endian;
     int arrays_seen = 0;
@@ -385,22 +254,22 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
     if (fields == NULL)
         return 0;
 
-    if (_PyObject_LookupAttr(type, &_Py_ID(_swappedbytes_), &tmp) < 0) {
+    int rc = PyObject_HasAttrWithError(type, &_Py_ID(_swappedbytes_));
+    if (rc < 0) {
         return -1;
     }
-    if (tmp) {
-        Py_DECREF(tmp);
+    if (rc) {
         big_endian = !PY_BIG_ENDIAN;
     }
     else {
         big_endian = PY_BIG_ENDIAN;
     }
 
-    if (_PyObject_LookupAttr(type, &_Py_ID(_pack_), &tmp) < 0) {
+    if (PyObject_GetOptionalAttr(type, &_Py_ID(_pack_), &tmp) < 0) {
         return -1;
     }
     if (tmp) {
-        pack = _PyLong_AsInt(tmp);
+        pack = PyLong_AsInt(tmp);
         Py_DECREF(tmp);
         if (pack < 0) {
             if (!PyErr_Occurred() ||
@@ -418,6 +287,28 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
         pack = 0;
     }
 
+    if (PyObject_GetOptionalAttr(type, &_Py_ID(_align_), &tmp) < 0) {
+        return -1;
+    }
+    if (tmp) {
+        forced_alignment = PyLong_AsInt(tmp);
+        Py_DECREF(tmp);
+        if (forced_alignment < 0) {
+            if (!PyErr_Occurred() ||
+                PyErr_ExceptionMatches(PyExc_TypeError) ||
+                PyErr_ExceptionMatches(PyExc_OverflowError))
+            {
+                PyErr_SetString(PyExc_ValueError,
+                                "_align_ must be a non-negative integer");
+            }
+            return -1;
+        }
+    }
+    else {
+        /* Setting `_align_ = 0` amounts to using the default alignment */
+        forced_alignment = 1;
+    }
+
     len = PySequence_Size(fields);
     if (len == -1) {
         if (PyErr_ExceptionMatches(PyExc_TypeError)) {
@@ -427,88 +318,97 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
         return -1;
     }
 
-    stgdict = PyType_stgdict(type);
-    if (!stgdict) {
+    ctypes_state *st = get_module_state_by_def(Py_TYPE(type));
+    StgInfo *stginfo;
+    if (PyStgInfo_FromType(st, type, &stginfo) < 0) {
+        return -1;
+    }
+    if (!stginfo) {
         PyErr_SetString(PyExc_TypeError,
                         "ctypes state is not initialized");
         return -1;
     }
+
     /* If this structure/union is already marked final we cannot assign
        _fields_ anymore. */
 
-    if (stgdict->flags & DICTFLAG_FINAL) {/* is final ? */
+    if (stginfo->flags & DICTFLAG_FINAL) {/* is final ? */
         PyErr_SetString(PyExc_AttributeError,
                         "_fields_ is final");
         return -1;
     }
 
-    if (stgdict->format) {
-        PyMem_Free(stgdict->format);
-        stgdict->format = NULL;
+    if (stginfo->format) {
+        PyMem_Free(stginfo->format);
+        stginfo->format = NULL;
     }
 
-    if (stgdict->ffi_type_pointer.elements)
-        PyMem_Free(stgdict->ffi_type_pointer.elements);
+    if (stginfo->ffi_type_pointer.elements)
+        PyMem_Free(stginfo->ffi_type_pointer.elements);
 
-    basedict = PyType_stgdict((PyObject *)((PyTypeObject *)type)->tp_base);
-    if (basedict) {
-        stgdict->flags |= (basedict->flags &
+    StgInfo *baseinfo;
+    if (PyStgInfo_FromType(st, (PyObject *)((PyTypeObject *)type)->tp_base,
+                           &baseinfo) < 0) {
+        return -1;
+    }
+    if (baseinfo) {
+        stginfo->flags |= (baseinfo->flags &
                            (TYPEFLAG_HASUNION | TYPEFLAG_HASBITFIELD));
     }
     if (!isStruct) {
-        stgdict->flags |= TYPEFLAG_HASUNION;
+        stginfo->flags |= TYPEFLAG_HASUNION;
     }
-    if (basedict) {
-        size = offset = basedict->size;
-        align = basedict->align;
+    if (baseinfo) {
+        size = offset = baseinfo->size;
+        align = baseinfo->align;
         union_size = 0;
         total_align = align ? align : 1;
-        stgdict->ffi_type_pointer.type = FFI_TYPE_STRUCT;
-        stgdict->ffi_type_pointer.elements = PyMem_New(ffi_type *, basedict->length + len + 1);
-        if (stgdict->ffi_type_pointer.elements == NULL) {
+        total_align = max(total_align, forced_alignment);
+        stginfo->ffi_type_pointer.type = FFI_TYPE_STRUCT;
+        stginfo->ffi_type_pointer.elements = PyMem_New(ffi_type *, baseinfo->length + len + 1);
+        if (stginfo->ffi_type_pointer.elements == NULL) {
             PyErr_NoMemory();
             return -1;
         }
-        memset(stgdict->ffi_type_pointer.elements, 0,
-               sizeof(ffi_type *) * (basedict->length + len + 1));
-        if (basedict->length > 0) {
-            memcpy(stgdict->ffi_type_pointer.elements,
-                   basedict->ffi_type_pointer.elements,
-                   sizeof(ffi_type *) * (basedict->length));
+        memset(stginfo->ffi_type_pointer.elements, 0,
+               sizeof(ffi_type *) * (baseinfo->length + len + 1));
+        if (baseinfo->length > 0) {
+            memcpy(stginfo->ffi_type_pointer.elements,
+                   baseinfo->ffi_type_pointer.elements,
+                   sizeof(ffi_type *) * (baseinfo->length));
         }
-        ffi_ofs = basedict->length;
+        ffi_ofs = baseinfo->length;
     } else {
         offset = 0;
         size = 0;
         align = 0;
         union_size = 0;
-        total_align = 1;
-        stgdict->ffi_type_pointer.type = FFI_TYPE_STRUCT;
-        stgdict->ffi_type_pointer.elements = PyMem_New(ffi_type *, len + 1);
-        if (stgdict->ffi_type_pointer.elements == NULL) {
+        total_align = forced_alignment == 0 ? 1 : forced_alignment;
+        stginfo->ffi_type_pointer.type = FFI_TYPE_STRUCT;
+        stginfo->ffi_type_pointer.elements = PyMem_New(ffi_type *, len + 1);
+        if (stginfo->ffi_type_pointer.elements == NULL) {
             PyErr_NoMemory();
             return -1;
         }
-        memset(stgdict->ffi_type_pointer.elements, 0,
+        memset(stginfo->ffi_type_pointer.elements, 0,
                sizeof(ffi_type *) * (len + 1));
         ffi_ofs = 0;
     }
 
-    assert(stgdict->format == NULL);
+    assert(stginfo->format == NULL);
     if (isStruct) {
-        stgdict->format = _ctypes_alloc_format_string(NULL, "T{");
+        stginfo->format = _ctypes_alloc_format_string(NULL, "T{");
     } else {
         /* PEP3118 doesn't support union. Use 'B' for bytes. */
-        stgdict->format = _ctypes_alloc_format_string(NULL, "B");
+        stginfo->format = _ctypes_alloc_format_string(NULL, "B");
     }
-    if (stgdict->format == NULL)
+    if (stginfo->format == NULL)
         return -1;
 
     for (i = 0; i < len; ++i) {
         PyObject *name = NULL, *desc = NULL;
         PyObject *pair = PySequence_GetItem(fields, i);
         PyObject *prop;
-        StgDictObject *dict;
         int bitsize = 0;
 
         if (!pair || !PyArg_ParseTuple(pair, "UO|i", &name, &desc, &bitsize)) {
@@ -517,24 +417,31 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
             Py_XDECREF(pair);
             return -1;
         }
-        if (PyCArrayTypeObject_Check(desc))
+        if (PyCArrayTypeObject_Check(st, desc)) {
             arrays_seen = 1;
-        dict = PyType_stgdict(desc);
-        if (dict == NULL) {
+        }
+
+        StgInfo *info;
+        if (PyStgInfo_FromType(st, desc, &info) < 0) {
+            Py_DECREF(pair);
+            return -1;
+        }
+        if (info == NULL) {
             Py_DECREF(pair);
             PyErr_Format(PyExc_TypeError,
                          "second item in _fields_ tuple (index %zd) must be a C type",
                          i);
             return -1;
         }
-        stgdict->ffi_type_pointer.elements[ffi_ofs + i] = &dict->ffi_type_pointer;
-        if (dict->flags & (TYPEFLAG_ISPOINTER | TYPEFLAG_HASPOINTER))
-            stgdict->flags |= TYPEFLAG_HASPOINTER;
-        stgdict->flags |= dict->flags & (TYPEFLAG_HASUNION | TYPEFLAG_HASBITFIELD);
-        dict->flags |= DICTFLAG_FINAL; /* mark field type final */
+
+        stginfo->ffi_type_pointer.elements[ffi_ofs + i] = &info->ffi_type_pointer;
+        if (info->flags & (TYPEFLAG_ISPOINTER | TYPEFLAG_HASPOINTER))
+            stginfo->flags |= TYPEFLAG_HASPOINTER;
+        stginfo->flags |= info->flags & (TYPEFLAG_HASUNION | TYPEFLAG_HASBITFIELD);
+        info->flags |= DICTFLAG_FINAL; /* mark field type final */
         if (PyTuple_Size(pair) == 3) { /* bits specified */
-            stgdict->flags |= TYPEFLAG_HASBITFIELD;
-            switch(dict->ffi_type_pointer.type) {
+            stginfo->flags |= TYPEFLAG_HASBITFIELD;
+            switch(info->ffi_type_pointer.type) {
             case FFI_TYPE_UINT8:
             case FFI_TYPE_UINT16:
             case FFI_TYPE_UINT32:
@@ -545,8 +452,8 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
             case FFI_TYPE_SINT8:
             case FFI_TYPE_SINT16:
             case FFI_TYPE_SINT32:
-                if (dict->getfunc != _ctypes_get_fielddesc("c")->getfunc
-                    && dict->getfunc != _ctypes_get_fielddesc("u")->getfunc
+                if (info->getfunc != _ctypes_get_fielddesc("c")->getfunc
+                    && info->getfunc != _ctypes_get_fielddesc("u")->getfunc
                     )
                     break;
                 /* else fall through */
@@ -557,7 +464,7 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
                 Py_DECREF(pair);
                 return -1;
             }
-            if (bitsize <= 0 || bitsize > dict->size * 8) {
+            if (bitsize <= 0 || bitsize > info->size * 8) {
                 PyErr_SetString(PyExc_ValueError,
                                 "number of bits invalid for bit field");
                 Py_DECREF(pair);
@@ -567,7 +474,7 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
             bitsize = 0;
 
         if (isStruct) {
-            const char *fieldfmt = dict->format ? dict->format : "B";
+            const char *fieldfmt = info->format ? info->format : "B";
             const char *fieldname = PyUnicode_AsUTF8(name);
             char *ptr;
             Py_ssize_t len;
@@ -583,7 +490,7 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
 
             /* construct the field now, as `prop->offset` is `offset` with
                corrected alignment */
-            prop = PyCField_FromDesc(desc, i,
+            prop = PyCField_FromDesc(st, desc, i,
                                    &field_size, bitsize, &bitofs,
                                    &size, &offset, &align,
                                    pack, big_endian);
@@ -597,10 +504,10 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
             padding = ((CFieldObject *)prop)->offset - last_size;
 
             if (padding > 0) {
-                ptr = stgdict->format;
-                stgdict->format = _ctypes_alloc_format_padding(ptr, padding);
+                ptr = stginfo->format;
+                stginfo->format = _ctypes_alloc_format_padding(ptr, padding);
                 PyMem_Free(ptr);
-                if (stgdict->format == NULL) {
+                if (stginfo->format == NULL) {
                     Py_DECREF(pair);
                     Py_DECREF(prop);
                     return -1;
@@ -618,17 +525,17 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
             }
             sprintf(buf, "%s:%s:", fieldfmt, fieldname);
 
-            ptr = stgdict->format;
-            if (dict->shape != NULL) {
-                stgdict->format = _ctypes_alloc_format_string_with_shape(
-                    dict->ndim, dict->shape, stgdict->format, buf);
+            ptr = stginfo->format;
+            if (info->shape != NULL) {
+                stginfo->format = _ctypes_alloc_format_string_with_shape(
+                    info->ndim, info->shape, stginfo->format, buf);
             } else {
-                stgdict->format = _ctypes_alloc_format_string(stgdict->format, buf);
+                stginfo->format = _ctypes_alloc_format_string(stginfo->format, buf);
             }
             PyMem_Free(ptr);
             PyMem_Free(buf);
 
-            if (stgdict->format == NULL) {
+            if (stginfo->format == NULL) {
                 Py_DECREF(pair);
                 Py_DECREF(prop);
                 return -1;
@@ -637,7 +544,7 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
             size = 0;
             offset = 0;
             align = 0;
-            prop = PyCField_FromDesc(desc, i,
+            prop = PyCField_FromDesc(st, desc, i,
                                    &field_size, bitsize, &bitofs,
                                    &size, &offset, &align,
                                    pack, big_endian);
@@ -663,6 +570,7 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
     }
 
     /* Adjust the size according to the alignment requirements */
+    assert(total_align != 0);
     aligned_size = ((size + total_align - 1) / total_align) * total_align;
 
     if (isStruct) {
@@ -672,38 +580,37 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
         /* Pad up to the full size of the struct */
         padding = aligned_size - size;
         if (padding > 0) {
-            ptr = stgdict->format;
-            stgdict->format = _ctypes_alloc_format_padding(ptr, padding);
+            ptr = stginfo->format;
+            stginfo->format = _ctypes_alloc_format_padding(ptr, padding);
             PyMem_Free(ptr);
-            if (stgdict->format == NULL) {
+            if (stginfo->format == NULL) {
                 return -1;
             }
         }
 
-        ptr = stgdict->format;
-        stgdict->format = _ctypes_alloc_format_string(stgdict->format, "}");
+        ptr = stginfo->format;
+        stginfo->format = _ctypes_alloc_format_string(stginfo->format, "}");
         PyMem_Free(ptr);
-        if (stgdict->format == NULL)
+        if (stginfo->format == NULL)
             return -1;
     }
 
-    stgdict->ffi_type_pointer.alignment = Py_SAFE_DOWNCAST(total_align,
+    stginfo->ffi_type_pointer.alignment = Py_SAFE_DOWNCAST(total_align,
                                                            Py_ssize_t,
                                                            unsigned short);
-    stgdict->ffi_type_pointer.size = aligned_size;
+    stginfo->ffi_type_pointer.size = aligned_size;
 
-    stgdict->size = aligned_size;
-    stgdict->align = total_align;
-    stgdict->length = len;      /* ADD ffi_ofs? */
+    stginfo->size = aligned_size;
+    stginfo->align = total_align;
+    stginfo->length = ffi_ofs + len;
 
 /*
- * On Arm platforms, structs with at most 4 elements of any floating point
- * type values can be passed through registers. If the type is double the
- * maximum size of the struct is 32 bytes.
- * By Arm platforms it is meant both 32 and 64-bit.
-*/
-#if defined(__aarch64__) || defined(__arm__)
+ * The value of MAX_STRUCT_SIZE depends on the platform Python is running on.
+ */
+#if defined(__aarch64__) || defined(__arm__) || defined(_M_ARM64)
 #  define MAX_STRUCT_SIZE 32
+#elif defined(__powerpc64__)
+#  define MAX_STRUCT_SIZE 64
 #else
 #  define MAX_STRUCT_SIZE 16
 #endif
@@ -714,24 +621,29 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
          * pointers, which is fine when an array name is being passed as
          * parameter, but not when passing structures by value that contain
          * arrays.
-         * On x86_64 Linux and Arm platforms, small structures passed by
-         * value are passed in registers, and in order to do this, libffi needs
-         * to know the true type of the array members of structs. Treating them
-         * as pointers breaks things.
+         * Small structures passed by value are passed in registers, and in
+         * order to do this, libffi needs to know the true type of the array
+         * members of structs. Treating them as pointers breaks things.
          *
-         * By small structures, we mean ones that are 16 bytes or less on
-         * x86-64 and 32 bytes or less on Arm. In that case, there can't be
-         * more than 16 or 32 elements after unrolling arrays, as we (will)
-         * disallow bitfields. So we can collect the true ffi_type values in
-         * a fixed-size local array on the stack and, if any arrays were seen,
-         * replace the ffi_type_pointer.elements with a more accurate set,
-         * to allow libffi to marshal them into registers correctly.
+         * Small structures have different sizes depending on the platform
+         * where Python is running on:
+         *
+         *      * x86-64: 16 bytes or less
+         *      * Arm platforms (both 32 and 64 bit): 32 bytes or less
+         *      * PowerPC 64 Little Endian: 64 bytes or less
+         *
+         * In that case, there can't be more than 16, 32 or 64 elements after
+         * unrolling arrays, as we (will) disallow bitfields.
+         * So we can collect the true ffi_type values in a fixed-size local
+         * array on the stack and, if any arrays were seen, replace the
+         * ffi_type_pointer.elements with a more accurate set, to allow
+         * libffi to marshal them into registers correctly.
          * It means one more loop over the fields, but if we got here,
          * the structure is small, so there aren't too many of those.
          *
-         * Although the passing in registers is specific to x86_64 Linux
-         * and Arm platforms, the array-in-struct vs. pointer problem is
-         * general. But we restrict the type transformation to small structs
+         * Although the passing in registers is specific to the above
+         * platforms, the array-in-struct vs. pointer problem is general.
+         * But we restrict the type transformation to small structs
          * nonetheless.
          *
          * Note that although a union may be small in terms of memory usage, it
@@ -758,8 +670,7 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
          *     struct { uint_32 e1; uint_32 e2; ... uint_32 e_4; } f6;
          * }
          *
-         * The same principle applies for a struct 32 bytes in size like in
-         * the case of Arm platforms.
+         * The same principle applies for a struct 32 or 64 bytes in size.
          *
          * So the struct/union needs setting up as follows: all non-array
          * elements copied across as is, and all array elements replaced with
@@ -782,7 +693,6 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
         for (i = 0; i < len; ++i) {
             PyObject *name, *desc;
             PyObject *pair = PySequence_GetItem(fields, i);
-            StgDictObject *dict;
             int bitsize = 0;
 
             if (pair == NULL) {
@@ -794,25 +704,34 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
                 Py_DECREF(pair);
                 return -1;
             }
-            dict = PyType_stgdict(desc);
-            if (dict == NULL) {
+
+            StgInfo *info;
+            if (PyStgInfo_FromType(st, desc, &info) < 0) {
+                Py_DECREF(pair);
+                return -1;
+            }
+            if (info == NULL) {
                 Py_DECREF(pair);
                 PyErr_Format(PyExc_TypeError,
                     "second item in _fields_ tuple (index %zd) must be a C type",
                     i);
                 return -1;
             }
-            if (!PyCArrayTypeObject_Check(desc)) {
+
+            if (!PyCArrayTypeObject_Check(st, desc)) {
                 /* Not an array. Just need an ffi_type pointer. */
                 num_ffi_type_pointers++;
             }
             else {
                 /* It's an array. */
-                Py_ssize_t length = dict->length;
-                StgDictObject *edict;
+                Py_ssize_t length = info->length;
 
-                edict = PyType_stgdict(dict->proto);
-                if (edict == NULL) {
+                StgInfo *einfo;
+                if (PyStgInfo_FromType(st, info->proto, &einfo) < 0) {
+                    Py_DECREF(pair);
+                    return -1;
+                }
+                if (einfo == NULL) {
                     Py_DECREF(pair);
                     PyErr_Format(PyExc_TypeError,
                         "second item in _fields_ tuple (index %zd) must be a C type",
@@ -860,9 +779,9 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
         if (num_ffi_types > 0) {
             memset(structs, 0, num_ffi_types * sizeof(ffi_type));
         }
-        if (ffi_ofs && (basedict != NULL)) {
+        if (ffi_ofs && (baseinfo != NULL)) {
             memcpy(element_types,
-                basedict->ffi_type_pointer.elements,
+                baseinfo->ffi_type_pointer.elements,
                 ffi_ofs * sizeof(ffi_type *));
         }
         element_index = ffi_ofs;
@@ -871,7 +790,6 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
         for (i = 0; i < len; ++i) {
             PyObject *name, *desc;
             PyObject *pair = PySequence_GetItem(fields, i);
-            StgDictObject *dict;
             int bitsize = 0;
 
             if (pair == NULL) {
@@ -891,9 +809,16 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
                 PyMem_Free(type_block);
                 return -1;
             }
-            dict = PyType_stgdict(desc);
+
+            StgInfo *info;
+            if (PyStgInfo_FromType(st, desc, &info) < 0) {
+                Py_DECREF(pair);
+                PyMem_Free(type_block);
+                return -1;
+            }
+
             /* Possibly this check could be avoided, but see above comment. */
-            if (dict == NULL) {
+            if (info == NULL) {
                 Py_DECREF(pair);
                 PyMem_Free(type_block);
                 PyErr_Format(PyExc_TypeError,
@@ -901,17 +826,21 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
                              i);
                 return -1;
             }
+
             assert(element_index < (ffi_ofs + len)); /* will be used below */
-            if (!PyCArrayTypeObject_Check(desc)) {
+            if (!PyCArrayTypeObject_Check(st, desc)) {
                 /* Not an array. Just copy over the element ffi_type. */
-                element_types[element_index++] = &dict->ffi_type_pointer;
+                element_types[element_index++] = &info->ffi_type_pointer;
             }
             else {
-                Py_ssize_t length = dict->length;
-                StgDictObject *edict;
-
-                edict = PyType_stgdict(dict->proto);
-                if (edict == NULL) {
+                Py_ssize_t length = info->length;
+                StgInfo *einfo;
+                if (PyStgInfo_FromType(st, info->proto, &einfo) < 0) {
+                    Py_DECREF(pair);
+                    PyMem_Free(type_block);
+                    return -1;
+                }
+                if (einfo == NULL) {
                     Py_DECREF(pair);
                     PyMem_Free(type_block);
                     PyErr_Format(PyExc_TypeError,
@@ -920,15 +849,15 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
                     return -1;
                 }
                 element_types[element_index++] = &structs[struct_index];
-                structs[struct_index].size = length * edict->ffi_type_pointer.size;
-                structs[struct_index].alignment = edict->ffi_type_pointer.alignment;
+                structs[struct_index].size = length * einfo->ffi_type_pointer.size;
+                structs[struct_index].alignment = einfo->ffi_type_pointer.alignment;
                 structs[struct_index].type = FFI_TYPE_STRUCT;
                 structs[struct_index].elements = &dummy_types[dummy_index];
                 ++struct_index;
                 /* Copy over the element's type, length times. */
                 while (length > 0) {
                     assert(dummy_index < (num_ffi_type_pointers));
-                    dummy_types[dummy_index++] = &edict->ffi_type_pointer;
+                    dummy_types[dummy_index++] = &einfo->ffi_type_pointer;
                     length--;
                 }
                 assert(dummy_index < (num_ffi_type_pointers));
@@ -942,19 +871,19 @@ PyCStructUnionType_update_stgdict(PyObject *type, PyObject *fields, int isStruct
          * Replace the old elements with the new, taking into account
          * base class elements where necessary.
          */
-        assert(stgdict->ffi_type_pointer.elements);
-        PyMem_Free(stgdict->ffi_type_pointer.elements);
-        stgdict->ffi_type_pointer.elements = element_types;
+        assert(stginfo->ffi_type_pointer.elements);
+        PyMem_Free(stginfo->ffi_type_pointer.elements);
+        stginfo->ffi_type_pointer.elements = element_types;
     }
 
     /* We did check that this flag was NOT set above, it must not
        have been set until now. */
-    if (stgdict->flags & DICTFLAG_FINAL) {
+    if (stginfo->flags & DICTFLAG_FINAL) {
         PyErr_SetString(PyExc_AttributeError,
                         "Structure or union cannot contain itself");
         return -1;
     }
-    stgdict->flags |= DICTFLAG_FINAL;
+    stginfo->flags |= DICTFLAG_FINAL;
 
     return MakeAnonFields(type);
 }
